#include <signal.h>
#include <sys/types.h>
#include <utmp.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include "untamo.h"

extern char *ctime(), *strcpy();
char *ttyname();
time_t time();

static	onalarm();

static	char	*uid, *line;
static	int	kcount, htype, pgrp;

/*
 * zap -- disconnect the person logged on to tty "dev".
 * 	  makes use of ioctl(2) to get a new controlling
 *	  terminal and the infinitely clever vhangup(2)
 * 	  to disconnect it.  
 */
zap(him, type)
struct user *him;
int type;
{
	static char message[] = "\n\n\007Logged out by the system.\n";
	int dts, td;
	int vh;
	char *tname;

	/* save info for onalarm */

	uid = him->uid;
	line = him->line;
	htype = type;
	kcount = 1;

	/*
	 * close all the child's descriptors 
	 */

	setpgrp(0,0);
	dts = getdtablesize();
	for ( dts--; dts>=0; dts--) {
	    (void) close(dts);
	}
	signal(SIGHUP, SIG_IGN);
	signal(SIGALRM, onalarm);
	signal(SIGTERM, SIG_DFL);

	alarm(60);	/* don't take more than a minute */

	/*
	 * now tell him it's over, and disconnect.
	 */
	if (-1 == (td = open(him->line, O_RDWR, 0600))) {
	    Log("open: %s",sys_errlist[errno]);
	}
	if (td >= 0)  {
	        if (-1 == ioctl(td, TIOCSTART, (char *)0)) {
		    Log("ioctl (TIOCSTART): %s",sys_errlist[errno]);
		}		
		if (-1 == ioctl(td, TIOCGPGRP, &pgrp)) {
		    Log("ioctl (TIOCGPGRP): %s",sys_errlist[errno]);
		    exit(0);
		}
		if (-1 == killpg(pgrp,SIGCONT)) {
		    Log("killpg (SIGCONT): %s:",sys_errlist[errno]);
		}
		if (-1 == write(td, message, sizeof(message))) {
		    Log("write (td=%d): %s",td,sys_errlist[errno]);
		}
		if (-1 == fsync(td)) {
		    Log("fsync(td=%d):",td,sys_errlist[errno]);
		}
		sleep(2);
		if (-1 == ioctl(td, TIOCFLUSH, (char *)0)) {
			Log("ioctl (TIOCFLUSH): %s",sys_errlist[errno]);
		}
		tname = ttyname(td);
		if (0 == him->remote) {	/* dont drop dtr on rlogins */
		    if (-1 == ioctl(td, TIOCCDTR, (char *)0)) {
			Log("ioctl (TIOCCDTR): %s", sys_errlist[errno]);
		    }
		    sleep(2);	/* getty does it this way... */
		}
#ifdef ULTRIX
		if (-1 == (vh = killpg(pgrp,SIGHUP))) {
		    Log("killpg (SIGHUP): %s:",sys_errlist[errno]);
		}
#else
		vh = vhangup();
#endif ULTRIX
		if (-1 == close(td)) {
			Log("close(%d): %s",td, sys_errlist[errno]);
		}
		if (fl_debug) {
#ifdef ULTRIX
		    Log("%s on %s (fd=%d, ct=%s, pgrp=%d) for %s, killpg=%d",
#else
		    Log("%s on %s (fd=%d, ct=%s, pgrp=%d) for %s, vhangup=%d",
#endif ULTRIX
			him->uid, him->line, td, tname, pgrp,
			( (type & IS_MULT) ? "mult" 
			: (type & IS_IDLE ? "idle" : "sess" )), vh );
		} else {
#ifdef ULTRIX
		    Log("%s on %s, for %s, killpg returned %d",
#else
		    Log("%s on %s, for %s, vhangup returned %d",
#endif ULTRIX
			him->uid, him->line, 
			( (type & IS_MULT) ? "multiple" 
			: (type & IS_IDLE ? "idle" : "session" )), vh );
		}
	} else {
		Log("couldn't open %s on %s",
			him->line, him->uid);
	}
}

static
onalarm()
{
    if (kcount > 1) {
    	Log("couldn't logout %s on %s for %s (kill 9/exiting)", uid, line,
			( (htype & IS_MULT) ? "multiple" 
			: (htype & IS_IDLE ? "idle" : "session" )));
	if (-1 == killpg(pgrp,SIGKILL)) {
	    Log("killpg (SIGKILL): %s:",sys_errlist[errno]);
	}
	exit(1);
    }
    kcount++;
    killpg(pgrp, SIGHUP);
    alarm(60);
}


error(sb)
char *sb;
{
	Log(sb);
}
